/// Functions for performing cryptographic operations on the main data structures.
mod tpm;

use crate::{Decompose, SectoredBuf, SECTOR_SZ_DEFAULT};

use super::{
    fmt, io, BigArray, Block, Deserialize, Display, Epoch, Formatter, Hashable, Header, Owned,
    Path, Principal, Read, Sectored, Seek, Serialize, TryCompose, Write, Writecap,
};

use foreign_types::ForeignType;
use openssl::{
    encrypt::{Decrypter as OsslDecrypter, Encrypter as OsslEncrypter},
    error::ErrorStack,
    hash::{hash, Hasher, MessageDigest},
    pkey::{HasPrivate, HasPublic, PKey},
    rand::rand_bytes,
    rsa::{Padding as OpensslPadding, Rsa as OsslRsa},
    sign::{Signer as OsslSigner, Verifier as OsslVerifier},
    symm::{decrypt as openssl_decrypt, encrypt as openssl_encrypt, Cipher, Crypter, Mode},
};
use serde::{
    de::{self, DeserializeOwned, Deserializer, SeqAccess, Visitor},
    ser::{SerializeStruct, Serializer},
};
use btserde::{self, from_vec, to_vec, write_to};
use std::{
    convert::Infallible,
    io::{ErrorKind, SeekFrom},
    marker::PhantomData,
    num::TryFromIntError,
    str::FromStr,
};
use strum_macros::{Display, EnumDiscriminants, EnumString};
use zeroize::ZeroizeOnDrop;

/// Data that may or may not be encrypted.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub(crate) enum Cryptotext<T: Serialize> {
    /// The inner value of `T` is plaintext.
    Plain(T),
    /// The inner value is ciphertext.
    Cipher(Vec<u8>),
}

/// Errors that can occur during cryptographic operations.
#[derive(Debug)]
pub enum Error {
    NoReadCap,
    NoKeyAvailable,
    MissingPrivateKey,
    KeyVariantUnsupported,
    BlockNotEncrypted,
    InvalidHashFormat,
    InvalidSignature,
    IncorrectSize { expected: usize, actual: usize },
    IndexOutOfBounds { index: usize, limit: usize },
    IndivisibleSize { divisor: usize, actual: usize },
    InvalidOffset { actual: usize, limit: usize },
    HashCmpFailure,
    RootHashNotVerified,
    WritecapAuthzErr(WritecapAuthzErr),
    Serde(btserde::Error),
    Io(std::io::Error),
    Custom(Box<dyn std::fmt::Debug + Send + Sync>),
}

impl Error {
    fn custom<E: std::fmt::Debug + Send + Sync + 'static>(err: E) -> Self {
        Error::Custom(Box::new(err))
    }
}

impl Display for Error {
    fn fmt(&self, f: &mut Formatter<'_>) -> fmt::Result {
        match self {
            Error::NoReadCap => write!(f, "no readcap"),
            Error::NoKeyAvailable => write!(f, "no key available"),
            Error::MissingPrivateKey => write!(f, "private key was missing"),
            Error::KeyVariantUnsupported => write!(f, "unsupported key variant"),
            Error::BlockNotEncrypted => write!(f, "block was not encrypted"),
            Error::InvalidHashFormat => write!(f, "invalid format"),
            Error::InvalidSignature => write!(f, "invalid signature"),
            Error::IncorrectSize { expected, actual } => {
                write!(f, "expected size {} but got {}", expected, actual)
            }
            Error::IndexOutOfBounds { index, limit } => write!(
                f,
                "index {} is out of bounds, it must be strictly less than {}",
                index, limit
            ),
            Error::IndivisibleSize { divisor, actual } => write!(
                f,
                "expected a size which is divisible by {} but got {}",
                divisor, actual
            ),
            Error::InvalidOffset { actual, limit } => write!(
                f,
                "offset {} is out of bounds, it must be strictly less than {}",
                actual, limit
            ),
            Error::HashCmpFailure => write!(f, "hash data are not equal"),
            Error::RootHashNotVerified => write!(f, "root hash is not verified"),
            Error::WritecapAuthzErr(err) => err.fmt(f),
            Error::Serde(err) => err.fmt(f),
            Error::Io(err) => err.fmt(f),
            Error::Custom(cus) => cus.fmt(f),
        }
    }
}

impl std::error::Error for Error {}

impl From<WritecapAuthzErr> for Error {
    fn from(err: WritecapAuthzErr) -> Self {
        Error::WritecapAuthzErr(err)
    }
}

impl From<ErrorStack> for Error {
    fn from(error: ErrorStack) -> Error {
        Error::custom(error)
    }
}

impl From<btserde::Error> for Error {
    fn from(error: btserde::Error) -> Error {
        Error::Serde(error)
    }
}

impl From<std::io::Error> for Error {
    fn from(error: std::io::Error) -> Error {
        Error::Io(error)
    }
}

impl From<TryFromIntError> for Error {
    fn from(err: TryFromIntError) -> Self {
        Error::custom(err)
    }
}

impl From<Error> for io::Error {
    fn from(err: Error) -> Self {
        io::Error::new(io::ErrorKind::Other, err)
    }
}

impl From<crate::Error> for Error {
    fn from(err: crate::Error) -> Self {
        Error::custom(err)
    }
}

pub(crate) type Result<T> = std::result::Result<T, Error>;

/// Returns an array of the given length filled with cryptographically strong random data.
fn rand_array<const LEN: usize>() -> Result<[u8; LEN]> {
    let mut array = [0; LEN];
    rand_bytes(&mut array)?;
    Ok(array)
}

/// Returns a vector of the given length with with cryptographically strong random data.
fn rand_vec(len: usize) -> Result<Vec<u8>> {
    let mut vec = vec![0; len];
    rand_bytes(&mut vec)?;
    Ok(vec)
}

/// A cryptographic hash.
#[derive(Debug, PartialEq, Eq, Serialize, Deserialize, Hashable, Clone, EnumDiscriminants)]
#[strum_discriminants(derive(EnumString, Display, Serialize, Deserialize))]
#[strum_discriminants(name(HashKind))]
pub enum Hash {
    Sha2_256([u8; HashKind::Sha2_256.len()]),
    #[serde(with = "BigArray")]
    Sha2_512([u8; HashKind::Sha2_512.len()]),
}

impl Default for HashKind {
    fn default() -> HashKind {
        HashKind::Sha2_256
    }
}

impl HashKind {
    pub const fn len(self) -> usize {
        match self {
            HashKind::Sha2_256 => 32,
            HashKind::Sha2_512 => 64,
        }
    }

    pub fn digest<'a, I: Iterator<Item = &'a [u8]>>(self, dest: &mut [u8], parts: I) -> Result<()> {
        if dest.len() != self.len() {
            return Err(Error::IncorrectSize {
                expected: self.len(),
                actual: dest.len(),
            });
        }
        let mut hasher = Hasher::new(self.into())?;
        for part in parts {
            hasher.update(part)?;
        }
        let hash = hasher.finish()?;
        dest.copy_from_slice(&hash);
        Ok(())
    }
}

impl From<HashKind> for MessageDigest {
    fn from(kind: HashKind) -> Self {
        match kind {
            HashKind::Sha2_256 => MessageDigest::sha256(),
            HashKind::Sha2_512 => MessageDigest::sha512(),
        }
    }
}

impl Hash {
    /// The character that's used to separate a hash type from its value in its string
    /// representation.
    const HASH_SEP: char = ':';

    fn kind(&self) -> HashKind {
        HashKind::from(self)
    }
}

impl From<HashKind> for Hash {
    fn from(discriminant: HashKind) -> Hash {
        match discriminant {
            HashKind::Sha2_256 => Hash::Sha2_256([0u8; HashKind::Sha2_256.len()]),
            HashKind::Sha2_512 => Hash::Sha2_512([0u8; HashKind::Sha2_512.len()]),
        }
    }
}

impl AsRef<[u8]> for Hash {
    fn as_ref(&self) -> &[u8] {
        match self {
            Hash::Sha2_256(arr) => arr,
            Hash::Sha2_512(arr) => arr,
        }
    }
}

impl AsMut<[u8]> for Hash {
    fn as_mut(&mut self) -> &mut [u8] {
        match self {
            Hash::Sha2_256(arr) => arr,
            Hash::Sha2_512(arr) => arr,
        }
    }
}

impl TryFrom<&str> for Hash {
    type Error = Error;
    fn try_from(string: &str) -> Result<Hash> {
        let mut split: Vec<&str> = string.split(Self::HASH_SEP).collect();
        if split.len() != 2 {
            return Err(Error::InvalidHashFormat);
        };
        let second = split.pop().ok_or(Error::InvalidHashFormat)?;
        let first = split.pop().ok_or(Error::InvalidHashFormat)?;
        let mut hash = Hash::from(HashKind::from_str(first).map_err(|_| Error::InvalidHashFormat)?);
        base64_url::decode_to_slice(second, hash.as_mut()).map_err(|_| Error::InvalidHashFormat)?;
        Ok(hash)
    }
}

impl Display for Hash {
    fn fmt(&self, f: &mut fmt::Formatter<'_>) -> fmt::Result {
        let hash_type: HashKind = self.into();
        let hash_data = base64_url::encode(self.as_ref());
        write!(f, "{}{}{}", hash_type, Hash::HASH_SEP, hash_data)
    }
}

/// A cryptographic signature.
#[derive(Debug, PartialEq, Serialize, Deserialize, Clone)]
pub struct Signature {
    kind: Sign,
    data: Vec<u8>,
}

impl Signature {
    pub fn empty(kind: Sign) -> Signature {
        let data = vec![0; kind.key_len() as usize];
        Signature { kind, data }
    }

    pub fn copy_from(kind: Sign, from: &[u8]) -> Signature {
        let mut data = vec![0; kind.key_len() as usize];
        data.as_mut_slice().copy_from_slice(from);
        Signature { kind, data }
    }

    pub fn as_slice(&self) -> &[u8] {
        self.data.as_slice()
    }

    pub fn as_mut_slice(&mut self) -> &mut [u8] {
        self.data.as_mut_slice()
    }
}

impl AsRef<[u8]> for Signature {
    fn as_ref(&self) -> &[u8] {
        self.as_slice()
    }
}

impl AsMut<[u8]> for Signature {
    fn as_mut(&mut self) -> &mut [u8] {
        self.as_mut_slice()
    }
}

#[derive(Serialize, Deserialize)]
struct TaggedCiphertext<T: Serialize, U: Serialize> {
    aad: U,
    ciphertext: Cryptotext<T>,
    tag: Vec<u8>,
}

#[derive(EnumDiscriminants, ZeroizeOnDrop)]
#[strum_discriminants(name(AeadKeyKind))]
#[strum_discriminants(derive(Serialize, Deserialize))]
pub enum AeadKey {
    AesGcm256 {
        key: [u8; AeadKeyKind::AesGcm256.key_len()],
        iv: [u8; AeadKeyKind::AesGcm256.iv_len()],
    },
}

impl AeadKeyKind {
    const fn key_len(self) -> usize {
        match self {
            AeadKeyKind::AesGcm256 => 32,
        }
    }

    const fn iv_len(self) -> usize {
        match self {
            AeadKeyKind::AesGcm256 => 16,
        }
    }
}

fn array_from<const N: usize>(slice: &[u8]) -> Result<[u8; N]> {
    let slice_len = slice.len();
    if N != slice_len {
        return Err(Error::IncorrectSize {
            actual: slice_len,
            expected: N,
        });
    }
    let mut array = [0u8; N];
    array.copy_from_slice(slice);
    Ok(array)
}

impl AeadKey {
    fn new(kind: AeadKeyKind) -> Result<AeadKey> {
        match kind {
            AeadKeyKind::AesGcm256 => Ok(AeadKey::AesGcm256 {
                key: rand_array()?,
                iv: rand_array()?,
            }),
        }
    }

    fn copy_components(kind: AeadKeyKind, key_buf: &[u8], iv_buf: &[u8]) -> Result<AeadKey> {
        match kind {
            AeadKeyKind::AesGcm256 => Ok(AeadKey::AesGcm256 {
                key: array_from(key_buf)?,
                iv: array_from(iv_buf)?,
            }),
        }
    }

    fn encrypt<T: Serialize + DeserializeOwned, U: Serialize + DeserializeOwned>(
        &self,
        aad: U,
        plaintext: &T,
    ) -> Result<TaggedCiphertext<T, U>> {
        let (cipher, key, iv, mut tag) = match self {
            AeadKey::AesGcm256 { key, iv } => (
                Cipher::aes_256_gcm(),
                key.as_slice(),
                iv.as_slice(),
                vec![0u8; 16],
            ),
        };

        let aad_data = to_vec(&aad)?;
        let plaintext_buf = to_vec(&plaintext)?;
        let mut ciphertext = vec![0u8; plaintext_buf.len() + cipher.block_size()];
        let mut crypter = Crypter::new(cipher, Mode::Encrypt, key, Some(iv))?;
        crypter.aad_update(&aad_data)?;
        let mut count = crypter.update(&plaintext_buf, &mut ciphertext)?;
        count += crypter.finalize(&mut ciphertext[count..])?;
        ciphertext.truncate(count);
        crypter.get_tag(&mut tag)?;

        Ok(TaggedCiphertext {
            aad,
            ciphertext: Cryptotext::Cipher(ciphertext),
            tag,
        })
    }

    fn decrypt<T: Serialize + DeserializeOwned, U: Serialize + DeserializeOwned>(
        &self,
        tagged: &TaggedCiphertext<T, U>,
    ) -> Result<T> {
        let ciphertext = match &tagged.ciphertext {
            Cryptotext::Plain(_) => return Err(Error::custom("already decrypted")),
            Cryptotext::Cipher(ciphertext) => ciphertext,
        };

        let (cipher, key, iv) = match self {
            AeadKey::AesGcm256 { key, iv } => {
                (Cipher::aes_256_gcm(), key.as_slice(), iv.as_slice())
            }
        };

        let mut plaintext = vec![0u8; ciphertext.len() + cipher.block_size()];
        let mut crypter = Crypter::new(cipher, Mode::Decrypt, key, Some(iv))?;
        crypter.set_tag(&tagged.tag)?;
        let aad_buf = to_vec(&tagged.aad)?;
        crypter.aad_update(&aad_buf)?;
        let mut count = crypter.update(ciphertext, &mut plaintext)?;
        count += crypter.finalize(&mut plaintext[count..])?;
        plaintext.truncate(count);

        Ok(from_vec(&plaintext)?)
    }
}

#[derive(Debug, PartialEq, Serialize, Deserialize, Clone, EnumDiscriminants)]
#[strum_discriminants(name(SymKeyKind))]
pub enum SymKey {
    /// A key for the AES 256 cipher in Cipher Block Chaining mode. Note that this includes the
    /// initialization vector, so that a value of this variant contains all the information needed
    /// to fully initialize a cipher context.
    Aes256Cbc { key: [u8; 32], iv: [u8; 16] },
    /// A key for the AES 256 cipher in counter mode.
    Aes256Ctr { key: [u8; 32], iv: [u8; 16] },
}

struct SymParams<'a> {
    cipher: Cipher,
    key: &'a [u8],
    iv: Option<&'a [u8]>,
}

impl SymKey {
    pub(crate) fn generate(kind: SymKeyKind) -> Result<SymKey> {
        match kind {
            SymKeyKind::Aes256Cbc => Ok(SymKey::Aes256Cbc {
                key: rand_array()?,
                iv: rand_array()?,
            }),
            SymKeyKind::Aes256Ctr => Ok(SymKey::Aes256Ctr {
                key: rand_array()?,
                iv: rand_array()?,
            }),
        }
    }

    fn params(&self) -> SymParams {
        let (cipher, key, iv) = match self {
            SymKey::Aes256Cbc { key, iv } => (Cipher::aes_256_cbc(), key, Some(iv.as_slice())),
            SymKey::Aes256Ctr { key, iv } => (Cipher::aes_256_ctr(), key, Some(iv.as_slice())),
        };
        SymParams { cipher, key, iv }
    }

    fn block_size(&self) -> usize {
        let SymParams { cipher, .. } = self.params();
        cipher.block_size()
    }

    // The number of bytes that the plaintext expands by when encrypted.
    fn expansion_sz(&self) -> usize {
        match self {
            SymKey::Aes256Cbc { .. } => 16,
            SymKey::Aes256Ctr { .. } => 0,
        }
    }

    fn to_encrypter(&self) -> Result<Crypter> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(Crypter::new(cipher, Mode::Encrypt, key, iv)?)
    }

    fn to_decrypter(&self) -> Result<Crypter> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(Crypter::new(cipher, Mode::Decrypt, key, iv)?)
    }
}

impl Encrypter for SymKey {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(openssl_encrypt(cipher, key, iv, slice)?)
    }
}

impl Decrypter for SymKey {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let SymParams { cipher, key, iv } = self.params();
        Ok(openssl_decrypt(cipher, key, iv, slice)?)
    }
}

#[repr(u32)]
#[derive(Debug, Display, Clone, Copy, Serialize, Deserialize, PartialEq, Eq)]
pub enum KeyLen {
    Bits256 = 32,
    Bits512 = 64,
    Bits2048 = 256,
    Bits3072 = 384,
    Bits4096 = 512,
}

impl KeyLen {
    const fn bits(self) -> u32 {
        8 * self as u32
    }
}

pub trait Scheme:
    for<'de> Deserialize<'de> + Serialize + Copy + std::fmt::Debug + PartialEq + Into<Self::Kind>
{
    type Kind: Scheme;
    fn as_enum(self) -> SchemeKind;
    fn message_digest(&self) -> MessageDigest;
    fn padding(&self) -> Option<OpensslPadding>;
    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>>;
    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>>;
    fn generate(self) -> Result<AsymKeyPair<Self::Kind>>;
    fn key_len(self) -> KeyLen;
}

pub enum SchemeKind {
    Sign(Sign),
    Encrypt(Encrypt),
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub enum Encrypt {
    RsaEsOaep(RsaEsOaep),
}

impl Scheme for Encrypt {
    type Kind = Encrypt;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Encrypt(self)
    }

    fn message_digest(&self) -> MessageDigest {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.message_digest(),
        }
    }

    fn padding(&self) -> Option<OpensslPadding> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.padding(),
        }
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.public_from_der(der),
        }
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.private_from_der(der),
        }
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.generate(),
        }
    }

    fn key_len(self) -> KeyLen {
        match self {
            Encrypt::RsaEsOaep(inner) => inner.key_len(),
        }
    }
}

impl Encrypt {
    pub const RSA_OAEP_2048_SHA_256: Encrypt = Encrypt::RsaEsOaep(RsaEsOaep {
        key_len: KeyLen::Bits2048,
        hash_kind: HashKind::Sha2_256,
    });

    pub const RSA_OAEP_3072_SHA_256: Encrypt = Encrypt::RsaEsOaep(RsaEsOaep {
        key_len: KeyLen::Bits3072,
        hash_kind: HashKind::Sha2_256,
    });
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub enum Sign {
    RsaSsaPss(RsaSsaPss),
}

impl Scheme for Sign {
    type Kind = Sign;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Sign(self)
    }

    fn message_digest(&self) -> MessageDigest {
        match self {
            Sign::RsaSsaPss(inner) => inner.message_digest(),
        }
    }

    fn padding(&self) -> Option<OpensslPadding> {
        match self {
            Sign::RsaSsaPss(inner) => inner.padding(),
        }
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        match self {
            Sign::RsaSsaPss(inner) => inner.public_from_der(der),
        }
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        match self {
            Sign::RsaSsaPss(inner) => inner.private_from_der(der),
        }
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        match self {
            Sign::RsaSsaPss(inner) => inner.generate(),
        }
    }

    fn key_len(self) -> KeyLen {
        self.key_len_const()
    }
}

impl Sign {
    pub const RSA_PSS_2048_SHA_256: Sign = Sign::RsaSsaPss(RsaSsaPss {
        key_bits: KeyLen::Bits2048,
        hash_kind: HashKind::Sha2_256,
    });

    pub const RSA_PSS_3072_SHA_256: Sign = Sign::RsaSsaPss(RsaSsaPss {
        key_bits: KeyLen::Bits3072,
        hash_kind: HashKind::Sha2_256,
    });

    const fn key_len_const(self) -> KeyLen {
        match self {
            Sign::RsaSsaPss(inner) => inner.key_bits,
        }
    }
}

enum Rsa {}

impl Rsa {
    /// The default public exponent to use for generated RSA keys.
    const EXP: u32 = 65537; // 2**16 + 1

    fn generate<S: Scheme>(scheme: S) -> Result<AsymKeyPair<S>> {
        let key = OsslRsa::generate(scheme.key_len().bits())?;
        // TODO: Separating the keys this way seems inefficient. Investigate alternatives.
        let public_der = key.public_key_to_der()?;
        let private_der = key.private_key_to_der()?;
        let public = AsymKey::<Public, S>::new(scheme, &public_der)?;
        let private = AsymKey::<Private, S>::new(scheme, &private_der)?;
        Ok(AsymKeyPair { public, private })
    }
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub struct RsaEsOaep {
    key_len: KeyLen,
    hash_kind: HashKind,
}

impl Scheme for RsaEsOaep {
    type Kind = Encrypt;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Encrypt(self.into())
    }

    fn message_digest(&self) -> MessageDigest {
        self.hash_kind.into()
    }

    fn padding(&self) -> Option<OpensslPadding> {
        Some(OpensslPadding::PKCS1_OAEP)
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        Ok(PKey::public_key_from_der(der)?.conv_pub())
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        Ok(PKey::private_key_from_der(der)?.conv_priv())
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        Rsa::generate(self.into())
    }

    fn key_len(self) -> KeyLen {
        self.key_len
    }
}

impl From<RsaEsOaep> for Encrypt {
    fn from(scheme: RsaEsOaep) -> Self {
        Encrypt::RsaEsOaep(scheme)
    }
}

#[derive(Deserialize, Serialize, Clone, Debug, PartialEq, Copy)]
pub struct RsaSsaPss {
    key_bits: KeyLen,
    hash_kind: HashKind,
}

impl Scheme for RsaSsaPss {
    type Kind = Sign;

    fn as_enum(self) -> SchemeKind {
        SchemeKind::Sign(self.into())
    }

    fn message_digest(&self) -> MessageDigest {
        self.hash_kind.into()
    }

    fn padding(&self) -> Option<OpensslPadding> {
        Some(OpensslPadding::PKCS1_PSS)
    }

    fn public_from_der(self, der: &[u8]) -> Result<PKey<Public>> {
        Ok(PKey::public_key_from_der(der)?.conv_pub())
    }

    fn private_from_der(self, der: &[u8]) -> Result<PKey<Private>> {
        Ok(PKey::private_key_from_der(der)?.conv_priv())
    }

    fn generate(self) -> Result<AsymKeyPair<Self::Kind>> {
        Rsa::generate(self.into())
    }

    fn key_len(self) -> KeyLen {
        self.key_bits
    }
}

impl From<RsaSsaPss> for Sign {
    fn from(scheme: RsaSsaPss) -> Self {
        Sign::RsaSsaPss(scheme)
    }
}

/// Marker trait for the `Public` and `Private` key privacy types.
pub trait KeyPrivacy {}

/// Represents keys which can be shared freely.
#[derive(Clone, Debug)]
pub enum Public {}

impl KeyPrivacy for Public {}

unsafe impl HasPublic for Public {}

/// Represents keys which must be kept confidential.
pub enum Private {}

impl KeyPrivacy for Private {}

unsafe impl HasPrivate for Private {}

trait PKeyExt<T> {
    /// Converts a PKey<T> to a PKey<U>. This hack allows for converting between openssl's
    /// Public and Private types and ours.
    fn conv_pkey<U>(self) -> PKey<U>;

    /// Convert from openssl's Public type to `crypto::Public`.
    fn conv_pub(self) -> PKey<Public>;

    /// Convert from openssl's Private type to `crypto::Private`.
    fn conv_priv(self) -> PKey<Private>;
}

impl<T> PKeyExt<T> for PKey<T> {
    fn conv_pkey<U>(self) -> PKey<U> {
        let ptr = self.as_ptr();
        let new_pkey = unsafe { PKey::from_ptr(ptr) };
        std::mem::forget(self);
        new_pkey
    }

    fn conv_pub(self) -> PKey<Public> {
        self.conv_pkey()
    }

    fn conv_priv(self) -> PKey<Private> {
        self.conv_pkey()
    }
}

/// Represents any kind of asymmetric key.
#[derive(Debug, Clone)]
pub struct AsymKey<P: KeyPrivacy, S: Scheme> {
    scheme: S,
    pkey: PKey<P>,
}

pub type AsymKeyPub<S> = AsymKey<Public, S>;

impl<S: Scheme> AsymKey<Public, S> {
    pub(crate) fn new(scheme: S, der: &[u8]) -> Result<AsymKey<Public, S>> {
        let pkey = scheme.public_from_der(der)?;
        Ok(AsymKey { scheme, pkey })
    }
}

impl<S: Scheme> AsymKey<Private, S> {
    pub(crate) fn new(scheme: S, der: &[u8]) -> Result<AsymKey<Private, S>> {
        let pkey = scheme.private_from_der(der)?;
        Ok(AsymKey { scheme, pkey })
    }
}

impl<'de, S: Scheme> Deserialize<'de> for AsymKey<Public, S> {
    fn deserialize<D: Deserializer<'de>>(d: D) -> std::result::Result<Self, D::Error> {
        const FIELDS: &[&str] = &["scheme", "pkey"];

        struct StructVisitor<S: Scheme>(PhantomData<S>);

        impl<'de, S: Scheme> Visitor<'de> for StructVisitor<S> {
            type Value = AsymKey<Public, S>;

            fn expecting(&self, formatter: &mut fmt::Formatter) -> fmt::Result {
                formatter.write_fmt(format_args!("struct {}", stringify!(AsymKey)))
            }

            fn visit_seq<V: SeqAccess<'de>>(
                self,
                mut seq: V,
            ) -> std::result::Result<Self::Value, V::Error> {
                let scheme: S = seq
                    .next_element()?
                    .ok_or_else(|| de::Error::missing_field(FIELDS[0]))?;
                let der: Vec<u8> = seq
                    .next_element()?
                    .ok_or_else(|| de::Error::missing_field(FIELDS[1]))?;
                AsymKey::<Public, _>::new(scheme, der.as_slice()).map_err(de::Error::custom)
            }
        }

        d.deserialize_struct(stringify!(AsymKey), FIELDS, StructVisitor(PhantomData))
    }
}

impl<S: Scheme> Serialize for AsymKey<Public, S> {
    fn serialize<T: Serializer>(&self, s: T) -> std::result::Result<T::Ok, T::Error> {
        let mut struct_s = s.serialize_struct(stringify!(AsymKey), 2)?;
        struct_s.serialize_field("scheme", &self.scheme)?;
        let der = self.pkey.public_key_to_der().unwrap();
        struct_s.serialize_field("pkey", der.as_slice())?;
        struct_s.end()
    }
}

impl<S: Scheme> PartialEq for AsymKey<Public, S> {
    fn eq(&self, other: &Self) -> bool {
        self.scheme == other.scheme && self.pkey.public_eq(&other.pkey)
    }
}

impl Owned for AsymKey<Public, Sign> {
    fn owner_of_kind(&self, kind: HashKind) -> Principal {
        let der = self.pkey.public_key_to_der().unwrap();
        let bytes = hash(kind.into(), der.as_slice()).unwrap();
        let mut hash_buf = Hash::from(kind);
        hash_buf.as_mut().copy_from_slice(&bytes);
        Principal(hash_buf)
    }
}

impl Encrypter for AsymKey<Public, Encrypt> {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let mut encrypter = OsslEncrypter::new(&self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            encrypter.set_rsa_padding(padding)?;
        }
        {
            let Encrypt::RsaEsOaep(inner) = self.scheme;
            encrypter.set_rsa_oaep_md(inner.message_digest())?;
        }
        let buffer_len = encrypter.encrypt_len(slice)?;
        let mut ciphertext = vec![0; buffer_len];
        let ciphertext_len = encrypter.encrypt(slice, &mut ciphertext)?;
        ciphertext.truncate(ciphertext_len);
        Ok(ciphertext)
    }
}

impl Decrypter for AsymKey<Private, Encrypt> {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        let mut decrypter = OsslDecrypter::new(&self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            decrypter.set_rsa_padding(padding)?;
        }
        {
            let Encrypt::RsaEsOaep(inner) = self.scheme;
            decrypter.set_rsa_oaep_md(inner.message_digest())?;
        }
        let buffer_len = decrypter.decrypt_len(slice)?;
        let mut plaintext = vec![0; buffer_len];
        let plaintext_len = decrypter.decrypt(slice, &mut plaintext)?;
        plaintext.truncate(plaintext_len);
        Ok(plaintext)
    }
}

impl Signer for AsymKey<Private, Sign> {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature> {
        let mut signer = OsslSigner::new(self.scheme.message_digest(), &self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            signer.set_rsa_padding(padding)?;
        }
        for part in parts {
            signer.update(part)?;
        }
        let mut signature = Signature::empty(self.scheme);
        signer.sign(signature.as_mut_slice())?;
        Ok(signature)
    }
}

impl Verifier for AsymKey<Public, Sign> {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()> {
        let mut verifier = OsslVerifier::new(self.scheme.message_digest(), &self.pkey)?;
        if let Some(padding) = self.scheme.padding() {
            verifier.set_rsa_padding(padding)?;
        }
        for part in parts {
            verifier.update(part)?;
        }
        if verifier.verify(signature)? {
            Ok(())
        } else {
            Err(Error::InvalidSignature)
        }
    }
}

pub struct AsymKeyPair<S: Scheme> {
    public: AsymKey<Public, S>,
    private: AsymKey<Private, S>,
}

impl<S: Scheme> AsymKeyPair<S> {
    pub fn new(scheme: S, public_der: &[u8], private_der: &[u8]) -> Result<AsymKeyPair<S>> {
        let public = AsymKey::<Public, _>::new(scheme, public_der)?;
        let private = AsymKey::<Private, _>::new(scheme, private_der)?;
        Ok(AsymKeyPair { public, private })
    }
}

impl Owned for AsymKeyPair<Sign> {
    fn owner_of_kind(&self, kind: HashKind) -> Principal {
        self.public.owner_of_kind(kind)
    }
}

impl Encrypter for AsymKeyPair<Encrypt> {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.public.encrypt(slice)
    }
}

impl Decrypter for AsymKeyPair<Encrypt> {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.private.decrypt(slice)
    }
}

impl Signer for AsymKeyPair<Sign> {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature> {
        self.private.sign(parts)
    }
}

impl Verifier for AsymKeyPair<Sign> {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()> {
        self.public.verify(parts, signature)
    }
}

pub struct ConcreteCreds {
    sign: AsymKeyPair<Sign>,
    encrypt: AsymKeyPair<Encrypt>,
}

impl ConcreteCreds {
    pub(crate) fn new(sign: AsymKeyPair<Sign>, encrypt: AsymKeyPair<Encrypt>) -> ConcreteCreds {
        ConcreteCreds { sign, encrypt }
    }

    #[cfg(test)]
    pub(crate) fn generate() -> Result<ConcreteCreds> {
        let encrypt = Encrypt::RSA_OAEP_3072_SHA_256.generate()?;
        let sign = Sign::RSA_PSS_3072_SHA_256.generate()?;
        Ok(ConcreteCreds { sign, encrypt })
    }
}

impl Verifier for ConcreteCreds {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()> {
        self.sign.verify(parts, signature)
    }
}

impl Encrypter for ConcreteCreds {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.encrypt.encrypt(slice)
    }
}

impl Owned for ConcreteCreds {
    fn owner_of_kind(&self, kind: HashKind) -> Principal {
        self.sign.owner_of_kind(kind)
    }
}

impl CredsPub for ConcreteCreds {
    fn public_sign(&self) -> &AsymKey<Public, Sign> {
        &self.sign.public
    }
}

impl Signer for ConcreteCreds {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature> {
        self.sign.sign(parts)
    }
}

impl Decrypter for ConcreteCreds {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>> {
        self.encrypt.decrypt(slice)
    }
}

impl CredsPriv for ConcreteCreds {}

impl Creds for ConcreteCreds {}

pub(crate) trait Encrypter {
    fn encrypt(&self, slice: &[u8]) -> Result<Vec<u8>>;
}

pub(crate) trait Decrypter {
    fn decrypt(&self, slice: &[u8]) -> Result<Vec<u8>>;
}

pub(crate) trait Signer {
    fn sign<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<Signature>;
}

pub(crate) trait Verifier {
    fn verify<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I, signature: &[u8]) -> Result<()>;
}

/// Trait for types which can be used as public credentials.
pub(crate) trait CredsPub: Verifier + Encrypter + Owned {
    /// Returns a reference to the public signing key which can be used to verify signatures.
    fn public_sign(&self) -> &AsymKey<Public, Sign>;
}

/// Trait for types which contain private credentials.
pub(crate) trait CredsPriv: Decrypter + Signer {}

/// Trait for types which contain both public and private credentials.
pub(crate) trait Creds: CredsPriv + CredsPub {}

/// A trait for types which store credentials.
pub(crate) trait CredStore {
    type CredHandle: Creds;
    type ExportedCreds: Serialize + for<'de> Deserialize<'de>;

    /// Returns the node credentials. If credentials haven't been generated, they are generated
    /// stored and returned.
    fn node_creds(&self) -> Result<Self::CredHandle>;
    /// Returns the root credentials. If no root credentials have been generated, or the provided
    /// password is incorrect, then an error is returned.
    fn root_creds(&self, password: &str) -> Result<Self::CredHandle>;
    /// Generates the root credentials and protects them using the given password. If the root
    /// credentials have already been generated then an error is returned.
    fn gen_root_creds(&self, password: &str) -> Result<Self::CredHandle>;
    fn storage_key(&self) -> Result<AsymKeyPub<Encrypt>>;
    fn export_root_creds(
        &self,
        root_creds: &Self::CredHandle,
        password: &str,
        new_parent: &AsymKeyPub<Encrypt>,
    ) -> Result<Self::ExportedCreds>;
    fn import_root_creds(
        &self,
        password: &str,
        exported: Self::ExportedCreds,
    ) -> Result<Self::CredHandle>;
}

/// Returns the base 2 logarithm of the given number. This function will return -1 when given 0, and
/// this is the only input for which a negative value is returned.
fn log2(mut n: usize) -> isize {
    // Is there a better implementation of this in std? I wasn't able to find an integer log2
    // function in std, so I wrote this naive implementation.
    if 0 == n {
        return -1;
    }
    let num_bits = 8 * std::mem::size_of::<usize>() as isize;
    for k in 0..num_bits {
        n >>= 1;
        if 0 == n {
            return k;
        }
    }
    num_bits
}

/// Returns 2^x. Note that 0 is returned for any negative input.
fn exp2(x: isize) -> usize {
    if x < 0 {
        0
    } else {
        1 << x
    }
}

/// Trait for types which can be used as nodes in a `MerkleTree`.
pub trait MerkleNode: Default + Serialize + for<'de> Deserialize<'de> {
    /// The kind of hash algorithm that this `HashData` uses.
    const KIND: HashKind;

    /// Creates a new `HashData` instance by hashing the data produced by the given iterator and
    /// storing it in self.
    fn new<'a, I: Iterator<Item = &'a [u8]>>(parts: I) -> Result<Self>;

    /// Combines the hash data from the given children and prefix and stores it in self. It is
    /// an error for no children to be provided (though one or the other may be `None`).
    fn combine<'a, I: Iterator<Item = &'a [u8]>>(
        &mut self,
        prefix: I,
        left: Option<&'a Self>,
        right: Option<&'a Self>,
    ) -> Result<()>;

    /// Returns `Ok(())` if self contains the given hash data, and `Err(Error::HashCmpFailure)`
    /// otherwise.
    fn assert_contains(&self, hash_data: &[u8]) -> Result<()>;

    /// Returns `Ok(())` if self contains the hash of the given data. Otherwise,
    /// `Err(Error::HashCmpFailure)` is returned.
    fn assert_contains_hash_of<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<()>;

    /// Returns `Ok(())` if the result of combining left and right is contained in self.
    fn assert_parent_of<'a, I: Iterator<Item = &'a [u8]>>(
        &self,
        prefix: I,
        left: Option<&'a Self>,
        right: Option<&'a Self>,
    ) -> Result<()>;

    /// Computes the hash of the data produced by the given iterator and writes it to the
    /// given slice.
    fn digest<'a, I: Iterator<Item = &'a [u8]>>(dest: &mut [u8], parts: I) -> Result<()> {
        Self::KIND.digest(dest, parts)
    }
}

// TODO: Once full const generic support lands we can use a HashKind as a const param. Then we won't
// need to have different structs to support different kinds of hashes.
/// A struct for storing SHA2 256 hashes in a `MerkleTree`.
#[derive(Default, Serialize, Deserialize)]
struct Sha2_256(Option<[u8; HashKind::Sha2_256.len()]>);

impl Sha2_256 {
    fn as_slice(&self) -> Option<&[u8]> {
        self.0.as_ref().map(|e| e.as_slice())
    }

    fn as_mut_slice(&mut self) -> Option<&mut [u8]> {
        self.0.as_mut().map(|e| e.as_mut_slice())
    }

    /// Returns a mutable reference to the array contained in self, if the array already exists.
    /// Otherwise, creates a new array filled with zeros owned by self and returns a
    /// reference.
    fn mut_or_init(&mut self) -> &mut [u8] {
        if self.0.is_none() {
            self.0 = Some([0; HashKind::Sha2_256.len()])
        }
        self.0.as_mut().unwrap()
    }

    // I think this is the most complicated function signature I've ever written in any language.
    /// Combines the given slices, together with the given prefix, and stores the resulting hash
    /// in `dest`. If neither `left` nor `right` is `Some`, then `when_neither` is called and
    /// whatever it returns is returned by this method.
    fn combine_hash_data<'a, I: Iterator<Item = &'a [u8]>, F: FnOnce() -> Result<()>>(
        dest: &mut [u8],
        prefix: I,
        left: Option<&'a [u8]>,
        right: Option<&'a [u8]>,
        when_neither: F,
    ) -> Result<()> {
        match (left, right) {
            (Some(left), Some(right)) => {
                Self::digest(dest, prefix.chain([left, right].into_iter()))
            }
            (Some(left), None) => Self::digest(dest, prefix.chain([left, b"None"].into_iter())),
            (None, Some(right)) => Self::digest(dest, prefix.chain([b"None", right].into_iter())),
            (None, None) => when_neither(),
        }
    }
}

impl MerkleNode for Sha2_256 {
    const KIND: HashKind = HashKind::Sha2_256;

    fn new<'a, I: Iterator<Item = &'a [u8]>>(parts: I) -> Result<Self> {
        let mut array = [0u8; Self::KIND.len()];
        Self::digest(&mut array, parts)?;
        Ok(Sha2_256(Some(array)))
    }

    fn combine<'a, I: Iterator<Item = &'a [u8]>>(
        &mut self,
        prefix: I,
        left: Option<&'a Self>,
        right: Option<&'a Self>,
    ) -> Result<()> {
        let left = left.and_then(|e| e.as_slice());
        let right = right.and_then(|e| e.as_slice());
        Self::combine_hash_data(self.mut_or_init(), prefix, left, right, || {
            Err(Error::custom(
                "at least one argument to combine needs to supply data",
            ))
        })
    }

    fn assert_contains(&self, hash_data: &[u8]) -> Result<()> {
        if let Some(slice) = self.as_slice() {
            if slice == hash_data {
                return Ok(());
            }
        }
        Err(Error::HashCmpFailure)
    }

    fn assert_contains_hash_of<'a, I: Iterator<Item = &'a [u8]>>(&self, parts: I) -> Result<()> {
        let mut buf = [0u8; Self::KIND.len()];
        Self::digest(&mut buf, parts)?;
        self.assert_contains(&buf)
    }

    fn assert_parent_of<'a, I: Iterator<Item = &'a [u8]>>(
        &self,
        prefix: I,
        left: Option<&'a Self>,
        right: Option<&'a Self>,
    ) -> Result<()> {
        let slice = match self.as_slice() {
            Some(slice) => slice,
            None => return Err(Error::HashCmpFailure),
        };
        let buf = {
            let mut buf = [0u8; Self::KIND.len()];
            let left = left.and_then(|e| e.as_slice());
            let right = right.and_then(|e| e.as_slice());
            Self::combine_hash_data(&mut buf, prefix, left, right, || {
                Err(Error::custom("logic error encountered"))
            })?;
            buf
        };
        if slice == buf {
            Ok(())
        } else {
            Err(Error::HashCmpFailure)
        }
    }
}

/// An index into a binary tree. This type provides convenience methods for navigating a tree.
#[derive(Copy, Clone, PartialEq, Eq, PartialOrd, Ord)]
struct BinTreeIndex(usize);

impl BinTreeIndex {
    /// Returns the index of the left child of this node.
    fn left(self) -> Self {
        Self(2 * self.0 + 1)
    }

    /// Returns the index of the right child of this node.
    fn right(self) -> Self {
        Self(2 * (self.0 + 1))
    }

    /// Returns the index of the parent of this node.
    fn parent(self) -> Option<Self> {
        if self.0 > 0 {
            Some(Self((self.0 - 1) / 2))
        } else {
            None
        }
    }

    /// Returns an iterator over the indices of all of this node's ancestors.
    fn ancestors(self) -> impl Iterator<Item = BinTreeIndex> {
        struct ParentIter(Option<BinTreeIndex>);

        impl Iterator for ParentIter {
            type Item = BinTreeIndex;

            fn next(&mut self) -> Option<Self::Item> {
                let parent = match self.0 {
                    Some(curr) => curr.parent(),
                    None => None,
                };
                self.0 = parent;
                parent
            }
        }

        ParentIter(Some(self))
    }
}

/// An implementation of a Merkle tree, a tree for storing hashes. This implementation is a binary
/// tree which stores its nodes in a vector to ensure data locality.
///
/// This type is used to provide integrity protection to a sequence of fixed sized units of data
/// called sectors. The size of the sectors are determined when the tree is created and cannot
/// be changed later. The hashes contained in the leaf nodes of this tree are hashes of sectors.
/// Each sector corresponds to an offset into the protected data, and in order to verify that a
/// sector has not been modified, you must supply the offset of the sector.
#[derive(Serialize, Deserialize)]
pub struct MerkleTree<T> {
    nodes: Vec<T>,
    /// The size of the sectors of data that this tree will protect.
    sector_sz: usize,
    #[serde(skip)]
    root_verified: bool,
}

impl<T> MerkleTree<T> {
    /// A byte to prefix data being hashed for leaf nodes. It's important that this is different
    /// from `INTERIOR_PREFIX`.
    const LEAF_PREFIX: &'static [u8] = b"Leaf";
    /// A byte to prefix data being hashed for interior nodes. It's important that this is different
    /// from 'LEAF_PREFIX`.
    const INTERIOR_PREFIX: &'static [u8] = b"Interior";

    /// Creates a new tree with no nodes in it and the given sector size.
    fn empty(sector_sz: usize) -> MerkleTree<T> {
        MerkleTree {
            nodes: Vec::new(),
            sector_sz,
            root_verified: true,
        }
    }

    /// Returns the number of generations in self. This method returns -1 when the tree is empty,
    /// and this is the only case where a negative value is returned.
    fn generations(&self) -> isize {
        log2(self.nodes.len())
    }

    /// Returns the number of nodes in a complete binary tree with the given number of
    /// generations. Note that `generations` is 0-based, so a tree with 1 node has 0 generations,
    /// and a tree with 3 has 1.
    fn len(generations: isize) -> usize {
        if generations >= 0 {
            exp2(generations + 1) - 1
        } else {
            0
        }
    }

    /// Returns the index of the last node in the tree.
    fn end(&self) -> BinTreeIndex {
        BinTreeIndex(self.nodes.len() - 1)
    }

    /// Returns a reference to the hash stored in the given node, or `Error::IndexOutOfBounds` if
    /// the given index doesn't exist.
    fn hash_at(&self, index: BinTreeIndex) -> Result<&T> {
        self.nodes.get(index.0).ok_or(Error::IndexOutOfBounds {
            index: index.0,
            limit: self.nodes.len(),
        })
    }

    /// Returns the index which corresponds to the given offset into the protected data.
    fn offset_to_index(&self, offset: usize) -> Result<BinTreeIndex> {
        let gens = self.generations();
        let sector_index = offset / self.sector_sz;
        let index_limit = exp2(gens);
        if sector_index >= index_limit {
            return Err(Error::InvalidOffset {
                actual: offset,
                limit: index_limit * self.sector_sz,
            });
        }
        Ok(BinTreeIndex(exp2(gens) - 1 + sector_index))
    }

    /// Returns an iterator of slices which need to be hashed along with the data to create a leaf
    /// node.
    fn leaf_parts(data: &[u8]) -> impl Iterator<Item = &[u8]> {
        [Self::LEAF_PREFIX, data].into_iter()
    }

    /// Returns an iterator of slices which need to be hashed along with the data to create an
    /// interior node.
    fn interior_prefix<'a>() -> impl Iterator<Item = &'a [u8]> {
        [Self::INTERIOR_PREFIX].into_iter()
    }
}

impl<T: MerkleNode> MerkleTree<T> {
    /// Checks that the root node contains the given hash data. If it does then `Ok(())` is
    /// returned. If it doesn't, then `Err(Error::HashCmpFailure)` is returned.
    fn assert_root_contains(&mut self, hash_data: &[u8]) -> Result<()> {
        match self.hash_at(BinTreeIndex(0)) {
            Ok(root) => {
                root.assert_contains(hash_data)?;
                self.root_verified = true;
                Ok(())
            }
            Err(_) => Err(Error::HashCmpFailure),
        }
    }

    /// Hashes the given data, adds a new node to the tree with its hash and updates the hashes
    /// of all parent nodes.
    fn write(&mut self, offset: usize, data: &[u8]) -> Result<()> {
        self.assert_sector_sz(data.len())?;

        let sector_index = offset / self.sector_sz;
        let generations = self.generations();
        let sector_index_sup = exp2(generations);
        if sector_index >= sector_index_sup {
            // Need to resize the tree.
            let generations_new = log2(sector_index) + 1;
            let new_cap = Self::len(generations_new) - self.nodes.len();
            self.nodes.reserve_exact(new_cap);
            // Extend the vector so there is enough room to fit the current leaves in the last
            // generation.
            let leaf_ct = self.nodes.len() - Self::len(generations - 1);
            let new_len = Self::len(generations_new - 1) + sector_index + 1;
            self.nodes.resize_with(new_len, T::default);
            // Shift all previously allocated nodes down the tree.
            let generation_gap = generations_new - generations;
            for gen in (0..(generations + 1)).rev() {
                let shift = exp2(gen + generation_gap) - exp2(gen);
                let start = exp2(gen) - 1;
                let end = start
                    + if gen == generations {
                        leaf_ct
                    } else {
                        exp2(gen)
                    };
                for index in start..end {
                    let new_index = index + shift;
                    self.nodes.swap(index, new_index);
                }
            }
            // Percolate up the old root to ensure that all nodes on the path from the old
            // root to the new root are initialized. This is not needed in the case where the
            // generation gap is only 1, as only the root is uninitialized in this case and it will
            // be initialized after inserting the new node below.
            if generation_gap > 1 && generations >= 0 {
                self.perc_up(BinTreeIndex(exp2(generation_gap) - 1))?;
            }
        }

        let index = self.offset_to_index(offset)?;
        if index.0 >= self.nodes.len() {
            self.nodes.resize_with(index.0 + 1, T::default);
        }
        self.nodes[index.0] = T::new(Self::leaf_parts(data))?;
        self.perc_up(index)
    }

    /// Percolates up the hash change to the given node to the root.
    fn perc_up(&mut self, start: BinTreeIndex) -> Result<()> {
        for index in start.ancestors() {
            self.combine_children(index)?;
        }
        Ok(())
    }

    /// Combines the hashes of the given node's children and stores it in the given node.
    fn combine_children(&mut self, index: BinTreeIndex) -> Result<()> {
        let left = index.left();
        let right = index.right();
        // Note that index < left && index < right.
        let split = index.0 + 1;
        let (front, back) = self.nodes.split_at_mut(split);
        let dest = &mut front[front.len() - 1];
        let left = back.get(left.0 - split);
        let right = back.get(right.0 - split);
        dest.combine(Self::interior_prefix(), left, right)
            .map_err(|_| Error::IndexOutOfBounds {
                index: index.0,
                limit: Self::len(self.generations() - 1),
            })
    }

    /// Verifies that the given data stored from the given offset into the protected data, has not
    /// been modified.
    fn verify(&self, offset: usize, data: &[u8]) -> Result<()> {
        if !self.root_verified {
            return Err(Error::RootHashNotVerified);
        }
        self.assert_sector_sz(data.len())?;
        let start = self.offset_to_index(offset)?;
        self.hash_at(start)?
            .assert_contains_hash_of(Self::leaf_parts(data))?;
        for index in start.ancestors() {
            let parent = self.hash_at(index)?;
            let left = self.hash_at(index.left()).ok();
            let right = self.hash_at(index.right()).ok();
            parent.assert_parent_of(Self::interior_prefix(), left, right)?;
        }
        Ok(())
    }
}

impl<T> Sectored for MerkleTree<T> {
    fn sector_sz(&self) -> usize {
        self.sector_sz
    }
}

struct MerkleStream<T, H> {
    inner: T,
    tree: MerkleTree<H>,
    offset: usize,
}

impl<H> MerkleStream<(), H> {
    fn new(tree: MerkleTree<H>) -> Self {
        MerkleStream {
            inner: (),
            tree,
            offset: 0,
        }
    }
}

impl<T, H> Sectored for MerkleStream<T, H> {
    fn sector_sz(&self) -> usize {
        self.tree.sector_sz()
    }
}

impl<T, H> TryCompose<T, MerkleStream<T, H>> for MerkleStream<(), H> {
    type Error = Infallible;
    fn try_compose(self, inner: T) -> std::result::Result<MerkleStream<T, H>, Infallible> {
        Ok(MerkleStream {
            inner,
            tree: self.tree,
            offset: self.offset,
        })
    }
}

impl<T, H> Decompose<T> for MerkleStream<T, H> {
    fn into_inner(self) -> T {
        self.inner
    }
}

impl<T: Write, H: MerkleNode> Write for MerkleStream<T, H> {
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        self.assert_sector_sz(buf.len())?;
        self.tree.write(self.offset, buf)?;
        let written = self.inner.write(buf)?;
        self.offset += self.sector_sz();
        Ok(written)
    }

    fn flush(&mut self) -> io::Result<()> {
        self.inner.flush()
    }
}

impl<T: Read, H: MerkleNode> Read for MerkleStream<T, H> {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        self.assert_sector_sz(buf.len())?;
        self.inner.read_exact(buf)?;
        self.tree.verify(self.offset, buf)?;
        self.offset += self.sector_sz();
        Ok(self.sector_sz())
    }
}

impl<T: Seek, H> Seek for MerkleStream<T, H> {
    fn seek(&mut self, pos: io::SeekFrom) -> io::Result<u64> {
        let from_start = self.inner.seek(pos)?;
        self.offset = from_start as usize;
        Ok(from_start)
    }
}

// A stream which encrypts all data written to it and decrypts all data read from it.
pub struct SecretStream<T> {
    inner: T,
    // The sector size of the inner stream. Reads and writes are only executed using buffers of
    // this size.
    inner_sect_sz: usize,
    // The sector size of this stream. Reads and writes are only accepted for buffers of this size.
    sect_sz: usize,
    key: SymKey,
    /// Buffer for ciphertext.
    ct_buf: Vec<u8>,
    /// Buffer for plaintext.
    pt_buf: Vec<u8>,
}

impl<T> SecretStream<T> {
    /// Given an offset into this stream, produces the corresponding offset into the inner stream.
    fn inner_offset(&self, outer_offset: u64) -> u64 {
        let sect_sz = self.sect_sz as u64;
        let inner_sect_sz = self.inner_sect_sz as u64;
        // We return the offset into the current sector, plus the size of all previous sectors.
        outer_offset % sect_sz + outer_offset / sect_sz * inner_sect_sz
    }

    /// Given an offset into the inner stream, returns the corresponding offset into this stream.
    fn outer_offset(&self, inner_offset: u64) -> u64 {
        let sect_sz = self.sect_sz as u64;
        let inner_sect_sz = self.inner_sect_sz as u64;
        inner_offset % inner_sect_sz + inner_offset / inner_sect_sz * sect_sz
    }
}

impl SecretStream<()> {
    pub fn new(key: SymKey) -> SecretStream<()> {
        SecretStream {
            inner: (),
            inner_sect_sz: 0,
            sect_sz: 0,
            key,
            ct_buf: Vec::new(),
            pt_buf: Vec::new(),
        }
    }
}

impl<T> Decompose<T> for SecretStream<T> {
    fn into_inner(self) -> T {
        self.inner
    }
}

impl<T, U: Sectored> TryCompose<U, SecretStream<U>> for SecretStream<T> {
    type Error = Error;
    fn try_compose(mut self, inner: U) -> Result<SecretStream<U>> {
        let inner_sect_sz = inner.sector_sz();
        let expansion_sz = self.key.expansion_sz();
        let sect_sz = inner_sect_sz - expansion_sz;
        let block_sz = self.key.block_size();
        if 0 != sect_sz % block_sz {
            return Err(Error::IndivisibleSize {
                divisor: block_sz,
                actual: sect_sz,
            });
        }
        self.pt_buf.resize(inner_sect_sz, 0);
        self.pt_buf.resize(inner_sect_sz + block_sz, 0);
        Ok(SecretStream {
            inner,
            inner_sect_sz,
            sect_sz: inner_sect_sz - expansion_sz,
            key: self.key,
            ct_buf: self.ct_buf,
            pt_buf: self.pt_buf,
        })
    }
}

impl<T> Sectored for SecretStream<T> {
    fn sector_sz(&self) -> usize {
        self.sect_sz
    }
}

impl<T: Write> Write for SecretStream<T> {
    fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
        self.assert_sector_sz(buf.len())?;

        self.ct_buf.resize(self.inner_sect_sz, 0);
        let mut encrypter = self.key.to_encrypter()?;
        let mut count = encrypter.update(buf, &mut self.ct_buf)?;
        count += encrypter.finalize(&mut self.ct_buf[count..])?;
        self.ct_buf.truncate(count);

        self.inner.write_all(&self.ct_buf).map(|_| buf.len())
    }

    fn flush(&mut self) -> io::Result<()> {
        self.inner.flush()
    }
}

impl<T: Read> Read for SecretStream<T> {
    fn read(&mut self, buf: &mut [u8]) -> io::Result<usize> {
        self.assert_sector_sz(buf.len())?;

        self.ct_buf.resize(self.inner_sect_sz, 0);
        match self.inner.read_exact(&mut self.ct_buf) {
            Ok(_) => (),
            Err(err) => {
                if err.kind() == io::ErrorKind::UnexpectedEof {
                    return Ok(0);
                } else {
                    return Err(err);
                }
            }
        }

        self.pt_buf
            .resize(self.inner_sect_sz + self.key.block_size(), 0);
        let mut decrypter = self.key.to_decrypter()?;
        let mut count = decrypter.update(&self.ct_buf, &mut self.pt_buf)?;
        count += decrypter.finalize(&mut self.pt_buf[count..])?;
        self.pt_buf.truncate(count);

        buf.copy_from_slice(&self.pt_buf);
        Ok(buf.len())
    }
}

impl<T: Seek> Seek for SecretStream<T> {
    fn seek(&mut self, pos: io::SeekFrom) -> io::Result<u64> {
        let outer_offset = match pos {
            SeekFrom::Start(offset) => offset,
            SeekFrom::Current(offset) => {
                let inner_offset = self.inner.stream_position()?;
                let outer_offset = self.outer_offset(inner_offset);
                if offset >= 0 {
                    outer_offset + offset as u64
                } else {
                    outer_offset - (-offset) as u64
                }
            }
            SeekFrom::End(_) => {
                // We can support this once stream_len is stabilized:
                // https://github.com/rust-lang/rust/issues/59359
                return Err(io::Error::new(
                    ErrorKind::Unsupported,
                    "seeking from the end of the stream is not supported",
                ));
            }
        };
        let inner_offset = self.inner_offset(outer_offset);
        self.inner.seek(SeekFrom::Start(inner_offset))?;
        Ok(outer_offset)
    }
}

pub(crate) fn unveil<T: Read + Write + Seek, C: Encrypter + Decrypter>(
    mut block: Block<T>,
    principal: &Principal,
    creds: &C,
) -> Result<Block<impl Read + Write + Seek>> {
    let (principal_owned, read_cap) = block
        .header
        .readcaps
        .remove_entry(principal)
        .ok_or(Error::NoReadCap)?;
    let block_key = decrypt(read_cap, creds)?;
    let mut block = block
        .compose_body(MerkleStream::new(MerkleTree::<Sha2_256>::empty(
            SECTOR_SZ_DEFAULT,
        )))
        .try_compose_body(SecretStream::new(block_key.clone()))?
        .try_compose_body(SectoredBuf::new())?;
    block
        .header
        .readcaps
        .insert(principal_owned, encrypt(&block_key, creds)?);
    Ok(block)
}

fn encrypt<T: Serialize, K: Encrypter>(value: &T, key: &K) -> Result<Cryptotext<T>> {
    let data = to_vec(value)?;
    let vec = key.encrypt(&data);
    Ok(Cryptotext::Cipher(vec?))
}

fn decrypt<T: Serialize + DeserializeOwned, K: Decrypter>(
    cryptotext: Cryptotext<T>,
    key: &K,
) -> Result<T> {
    let data = match cryptotext {
        Cryptotext::Plain(value) => return Ok(value),
        Cryptotext::Cipher(data) => data,
    };
    let vec = key.decrypt(&data);
    Ok(from_vec(&vec?)?)
}

pub(crate) fn sign_block<T, K: Signer>(block: &mut Block<T>, priv_key: &K) -> Result<()> {
    let header = to_vec(&block.header)?;
    let signature = priv_key.sign(std::iter::once(header.as_slice()))?;
    block.sig = signature;
    Ok(())
}

pub(crate) fn verify_block<T>(block: &Block<T>) -> Result<()> {
    verify_writecap(&block.header.writecap, &block.header.path)?;
    let header_data = to_vec(&block.header)?;
    block.header.writecap.signing_key.verify(
        std::iter::once(header_data.as_slice()),
        block.sig.as_slice(),
    )
}

#[derive(Serialize)]
struct WritecapSigInput<'a> {
    issued_to: &'a Principal,
    path: &'a Path,
    expires: &'a Epoch,
    signing_key: &'a AsymKey<Public, Sign>,
}

impl<'a> From<&'a Writecap> for WritecapSigInput<'a> {
    fn from(writecap: &'a Writecap) -> WritecapSigInput<'a> {
        WritecapSigInput {
            issued_to: &writecap.issued_to,
            path: &writecap.path,
            expires: &writecap.expires,
            signing_key: &writecap.signing_key,
        }
    }
}

pub(crate) fn sign_writecap<K: Signer>(writecap: &mut Writecap, priv_key: &K) -> Result<()> {
    let sig_input = to_vec(&WritecapSigInput::from(&*writecap))?;
    writecap.signature = priv_key.sign([sig_input.as_slice()].into_iter())?;
    Ok(())
}

/// The types of errors which can occur when verifying a writecap chain is authorized to write to
/// a given path.
#[derive(Debug, PartialEq, Display)]
pub enum WritecapAuthzErr {
    /// The chain is not valid for use on the given path.
    UnauthorizedPath,
    /// At least one writecap in the chain is expired.
    Expired,
    /// The given writecaps do not actually form a chain.
    NotChained,
    /// The principal the root writecap was issued to does not own the given path.
    RootDoesNotOwnPath,
    /// An error occurred while serializing a writecap.
    Serde(String),
    /// The write cap chain was too long to be validated. The value contained in this error is
    /// the maximum allowed length.
    ChainTooLong(usize),
}

/// Verifies that the given `Writecap` actually grants permission to write to the given `Path`.
pub(crate) fn verify_writecap(mut writecap: &Writecap, path: &Path) -> Result<()> {
    const CHAIN_LEN_LIMIT: usize = 256;
    let mut prev: Option<&Writecap> = None;
    let mut sig_input = Vec::new();
    let now = Epoch::now();
    for _ in 0..CHAIN_LEN_LIMIT {
        if !writecap.path.contains(path) {
            return Err(WritecapAuthzErr::UnauthorizedPath.into());
        }
        if writecap.expires <= now {
            return Err(WritecapAuthzErr::Expired.into());
        }
        if let Some(prev) = &prev {
            if prev.signing_key.owner_of_kind(writecap.issued_to.kind()) != writecap.issued_to {
                return Err(WritecapAuthzErr::NotChained.into());
            }
        }
        let sig = WritecapSigInput::from(writecap);
        sig_input.clear();
        write_to(&sig, &mut sig_input).map_err(|e| WritecapAuthzErr::Serde(e.to_string()))?;
        writecap.signing_key.verify(
            [sig_input.as_slice()].into_iter(),
            writecap.signature.as_slice(),
        )?;
        match &writecap.next {
            Some(next) => {
                prev = Some(writecap);
                writecap = next;
            }
            None => {
                // We're at the root key. As long as the signer of this writecap is the owner of
                // the path, then the writecap is valid.
                if writecap.signing_key.owner_of_kind(path.owner.kind()) == path.owner {
                    return Ok(());
                } else {
                    return Err(WritecapAuthzErr::RootDoesNotOwnPath.into());
                }
            }
        }
    }
    Err(WritecapAuthzErr::ChainTooLong(CHAIN_LEN_LIMIT).into())
}

pub fn verify_header(_header: &Header, _sig: &Signature) -> Result<()> {
    unimplemented!()
}

#[cfg(test)]
mod tests {
    use super::*;
    use crate::{test_helpers::*, BrotliParams, SectoredBuf, SECTOR_SZ_DEFAULT};
    use std::{
        io::{Cursor, SeekFrom},
        time::Duration,
    };

    fn encrypt_decrypt_block_test_case<T: Read + Write + Seek, C: Creds>(
        mut block: Block<T>,
        principal: &Principal,
        creds: &C,
    ) {
        let expected = {
            let mut dest = Vec::new();
            block.seek(SeekFrom::Start(0)).expect("seek failed");
            block
                .body
                .read_to_end(&mut dest)
                .expect("read_to_end failed");
            block.seek(SeekFrom::Start(0)).expect("seek failed");
            dest
        };
        let mut block = unveil(block, principal, creds).expect("unveil failed");
        block.write_all(&expected).expect("write_all failed");
        block.flush().expect("flush failed");
        block.seek(SeekFrom::Start(0)).expect("seek failed");
        let actual = {
            let mut actual = Vec::new();
            block
                .read_to_end(&mut actual)
                .expect("failed to read actual block contents");
            actual
        };
        assert_eq!(expected, actual);
    }

    #[test]
    fn encrypt_decrypt_block() {
        let readcap = make_readcap();
        let principal = readcap.issued_to.clone();
        let block = make_block_with(readcap);
        let key = make_key_pair();
        encrypt_decrypt_block_test_case(block, &principal, &key)
    }

    #[test]
    fn rsa_sign_and_verify() -> Result<()> {
        let key = make_key_pair();
        let header = b"About: lyrics".as_slice();
        let message = b"Everything that feels so good is bad bad bad.".as_slice();
        let signature = key.sign([header, message].into_iter())?;
        key.verify([header, message].into_iter(), signature.as_slice())
    }

    #[test]
    fn sign_verify_block_rsa() -> Result<()> {
        let readcap = make_readcap();
        let principal = readcap.issued_to.clone();
        let block = make_block_with(readcap);
        let key = make_key_pair();
        let mut block = unveil(block, &principal, &key).expect("unveil failed");
        sign_block(&mut block, &key)?;
        verify_block(&block)?;
        Ok(())
    }

    #[test]
    fn hash_to_string() {
        let hash = make_principal().0;
        let string = hash.to_string();
        assert_eq!(
            "Sha2_256:dSip4J0kurN5VhVo_aTipM-ywOOWrqJuRRVQ7aa-bew",
            string
        )
    }

    #[test]
    fn hash_to_string_round_trip() -> Result<()> {
        let expected = make_principal().0;
        let string = expected.to_string();
        let actual = Hash::try_from(string.as_str())?;
        assert_eq!(expected, actual);
        Ok(())
    }

    #[test]
    fn verify_writecap_valid() {
        let writecap = make_writecap(vec!["apps", "verse"]);
        verify_writecap(&writecap, &writecap.path).expect("failed to verify writecap");
    }

    #[test]
    fn verify_writecap_invalid_signature() -> Result<()> {
        let mut writecap = make_writecap(vec!["apps", "verse"]);
        writecap.signature = Signature::empty(Sign::RSA_PSS_3072_SHA_256);
        let result = verify_writecap(&writecap, &writecap.path);
        if let Err(Error::InvalidSignature) = result {
            Ok(())
        } else {
            Err(Error::custom(format!("unexpected result {:?}", result)))
        }
    }

    fn assert_authz_err<T: std::fmt::Debug>(
        expected: WritecapAuthzErr,
        result: Result<T>,
    ) -> Result<()> {
        if let Err(Error::WritecapAuthzErr(actual)) = &result {
            if *actual == expected {
                return Ok(());
            }
        }
        Err(Error::custom(format!("unexpected result: {:?}", result)))
    }

    #[test]
    fn verify_writecap_invalid_path_not_contained() -> Result<()> {
        let writecap = make_writecap(vec!["apps", "verse"]);
        let mut path = writecap.path.clone();
        path.components.pop();
        // `path` is now a superpath of `writecap.path`, thus the writecap is not authorized to
        // write to it.
        let result = verify_writecap(&writecap, &path);
        assert_authz_err(WritecapAuthzErr::UnauthorizedPath, result)
    }

    #[test]
    fn verify_writecap_invalid_expired() -> Result<()> {
        let mut writecap = make_writecap(vec!["apps", "verse"]);
        writecap.expires = Epoch::now() - Duration::from_secs(1);
        let result = verify_writecap(&writecap, &writecap.path);
        assert_authz_err(WritecapAuthzErr::Expired, result)
    }

    #[test]
    fn verify_writecap_invalid_not_chained() -> Result<()> {
        let (mut root_writecap, root_key) = make_self_signed_writecap();
        root_writecap.issued_to = Principal(Hash::Sha2_256([0; 32]));
        sign_writecap(&mut root_writecap, &root_key)?;
        let node_key = AsymKeyPub::new(Sign::RSA_PSS_3072_SHA_256, NODE_PUBLIC_KEY.as_slice())?;
        let node_principal = node_key.owner();
        let writecap = make_writecap_trusted_by(
            root_writecap,
            root_key,
            node_principal,
            vec!["apps", "contacts"],
        );
        let result = verify_writecap(&writecap, &writecap.path);
        assert_authz_err(WritecapAuthzErr::NotChained, result)
    }

    #[test]
    fn verify_writecap_invalid_root_doesnt_own_path() -> Result<()> {
        let (mut root_writecap, root_key) = make_self_signed_writecap();
        let owner = Principal(Hash::Sha2_256([0; 32]));
        root_writecap.path = make_path_with_owner(owner, vec![]);
        sign_writecap(&mut root_writecap, &root_key)?;
        let node_key = AsymKeyPub::new(Sign::RSA_PSS_3072_SHA_256, NODE_PUBLIC_KEY.as_slice())?;
        let node_owner = node_key.owner();
        let writecap = make_writecap_trusted_by(
            root_writecap,
            root_key,
            node_owner,
            vec!["apps", "contacts"],
        );
        let result = verify_writecap(&writecap, &writecap.path);
        assert_authz_err(WritecapAuthzErr::RootDoesNotOwnPath, result)
    }

    #[test]
    fn aeadkey_encrypt_decrypt_aes256gcm() {
        let key = AeadKey::new(AeadKeyKind::AesGcm256).expect("failed to create key");
        let aad = [0u8; 16];
        let expected = [0u8; 32];
        let tagged = key.encrypt(aad, &expected).expect("encrypt failed");
        let actual = key.decrypt(&tagged).expect("decrypt failed");
        assert_eq!(expected, actual.as_slice());
    }

    #[test]
    fn aeadkey_decrypt_fails_when_ct_modified() {
        let key = AeadKey::new(AeadKeyKind::AesGcm256).expect("failed to create key");
        let aad = [0u8; 16];
        let expected = [0u8; 32];
        let mut tagged = key.encrypt(aad, &expected).expect("encrypt failed");
        {
            let ct = match &mut tagged.ciphertext {
                Cryptotext::Cipher(ct) => Some(ct),
                Cryptotext::Plain(_) => None,
            }
            .expect("failed to extract cipher text");
            ct[0] += 1;
        }
        let result = key.decrypt(&tagged);
        assert!(result.is_err())
    }

    fn secret_stream_sequential_test_case(key: SymKey, inner_sect_sz: usize, sect_ct: usize) {
        let mut stream = SecretStream::new(key)
            .try_compose(SectoredCursor::new(
                vec![0u8; inner_sect_sz * sect_ct],
                inner_sect_sz,
            ))
            .expect("compose failed");
        let sector_sz = stream.sector_sz();
        for k in 0..sect_ct {
            let sector = vec![k as u8; sector_sz];
            stream.write(&sector).expect("write failed");
        }
        stream.seek(SeekFrom::Start(0)).expect("seek failed");
        for k in 0..sect_ct {
            let expected = vec![k as u8; sector_sz];
            let mut actual = vec![0u8; sector_sz];
            stream.read(&mut actual).expect("read failed");
            assert_eq!(expected, actual);
        }
    }

    fn secret_stream_sequential_test_suite(kind: SymKeyKind) {
        let key = SymKey::generate(kind).expect("key generation failed");
        secret_stream_sequential_test_case(key.clone(), SECTOR_SZ_DEFAULT, 16);
    }

    #[test]
    fn secret_stream_encrypt_decrypt_are_inverse_aes256cbc() {
        secret_stream_sequential_test_suite(SymKeyKind::Aes256Cbc)
    }

    #[test]
    fn secret_stream_encrypt_decrypt_are_inverse_aes256ctr() {
        secret_stream_sequential_test_suite(SymKeyKind::Aes256Ctr)
    }

    fn secret_stream_random_access_test_case(
        rando: Randomizer,
        key: SymKey,
        inner_sect_sz: usize,
        sect_ct: usize,
    ) {
        let mut stream = SecretStream::new(key)
            .try_compose(SectoredCursor::new(
                vec![0u8; inner_sect_sz * sect_ct],
                inner_sect_sz,
            ))
            .expect("compose failed");
        let sect_sz = stream.sector_sz();
        let indices: Vec<usize> = rando.take(sect_ct).map(|e| e % sect_ct).collect();
        for index in indices.iter().map(|e| *e) {
            let offset = index * sect_sz;
            stream
                .seek(SeekFrom::Start(offset as u64))
                .expect("seek to write failed");
            let sector = vec![index as u8; sect_sz];
            stream.write(&sector).expect("write failed");
        }
        for index in indices.iter().map(|e| *e) {
            let offset = index * sect_sz;
            stream
                .seek(SeekFrom::Start(offset as u64))
                .expect("seek to read failed");
            let expected = vec![index as u8; sect_sz];
            let mut actual = vec![0u8; sect_sz];
            stream.read(&mut actual).expect("read failed");
            assert_eq!(expected, actual);
        }
    }

    fn secret_stream_random_access_test_suite(kind: SymKeyKind) {
        const SEED: [u8; Randomizer::HASH.len()] = [3u8; Randomizer::HASH.len()];
        let key = SymKey::generate(kind).expect("key generation failed");
        secret_stream_random_access_test_case(
            Randomizer::new(SEED),
            key.clone(),
            SECTOR_SZ_DEFAULT,
            20,
        );
        secret_stream_random_access_test_case(
            Randomizer::new(SEED),
            key.clone(),
            SECTOR_SZ_DEFAULT,
            800,
        );
        secret_stream_random_access_test_case(Randomizer::new(SEED), key.clone(), 512, 200);
        secret_stream_random_access_test_case(Randomizer::new(SEED), key.clone(), 512, 20);
        secret_stream_random_access_test_case(Randomizer::new(SEED), key.clone(), 512, 200);
    }

    #[test]
    fn secret_stream_random_access() {
        secret_stream_random_access_test_suite(SymKeyKind::Aes256Cbc);
        secret_stream_random_access_test_suite(SymKeyKind::Aes256Ctr);
    }

    fn make_secret_stream(
        key_kind: SymKeyKind,
        num_sectors: usize,
    ) -> SecretStream<SectoredCursor<Vec<u8>>> {
        let key = SymKey::generate(key_kind).expect("key generation failed");
        let inner = SectoredCursor::new(
            vec![0u8; num_sectors * SECTOR_SZ_DEFAULT],
            SECTOR_SZ_DEFAULT,
        );
        SecretStream::new(key)
            .try_compose(inner)
            .expect("compose failed")
    }

    #[test]
    fn secret_stream_seek_from_start() {
        let mut stream = make_secret_stream(SymKeyKind::Aes256Cbc, 3);
        let sector_sz = stream.sector_sz();
        let expected = vec![2u8; sector_sz];
        // Write one sector of ones, one sector of twos and one sector of threes.
        for k in 1..4 {
            let sector: Vec<u8> = std::iter::repeat(k as u8).take(sector_sz).collect();
            stream.write(&sector).expect("writing to stream failed");
            fn secret_stream(
                key_kind: SymKeyKind,
                num_sectors: usize,
            ) -> SecretStream<SectoredCursor<Vec<u8>>> {
                let key = SymKey::generate(key_kind).expect("key generation failed");
                let inner = SectoredCursor::new(
                    vec![0u8; num_sectors * SECTOR_SZ_DEFAULT],
                    SECTOR_SZ_DEFAULT,
                );
                SecretStream::new(key)
                    .try_compose(inner)
                    .expect("compose failed")
            }
        }

        stream
            .seek(SeekFrom::Start(sector_sz as u64))
            .expect("seek failed");

        // A read from the stream should now return the second sector, which is filled with twos.
        let mut actual = vec![0u8; sector_sz];
        stream
            .read(&mut actual)
            .expect("reading from stream failed");
        assert_eq!(expected, actual);
    }

    #[test]
    fn secret_stream_seek_from_current() {
        let mut stream = make_secret_stream(SymKeyKind::Aes256Cbc, 3);
        let sector_sz = stream.sector_sz();
        let expected = vec![3u8; sector_sz];
        // Write one sector of ones, one sector of twos and one sector of threes.
        for k in 1..4 {
            let sector: Vec<u8> = std::iter::repeat(k as u8).take(sector_sz).collect();
            stream.write(&sector).expect("writing to stream failed");
        }

        stream
            .seek(SeekFrom::Current(-1 * (sector_sz as i64)))
            .expect("seek failed");

        // A read from the stream should now return the last sector, which is filled with threes.
        let mut actual = vec![0u8; sector_sz];
        stream
            .read(&mut actual)
            .expect("reading from stream failed");
        assert_eq!(expected, actual);
    }

    #[test]
    fn log2_test() {
        assert_eq!(-1, log2(0));
        assert_eq!(0, log2(1));
        assert_eq!(1, log2(2));
        assert_eq!(2, log2(4));
        assert_eq!(2, log2(5));
        assert_eq!(3, log2(8));
        assert_eq!(9, log2(1023));
        assert_eq!(10, log2(1025));
        assert_eq!(63, log2(usize::MAX));
    }

    fn make_tree_with<const SZ: usize>(num_sects: usize) -> (MerkleTree<Sha2_256>, Vec<[u8; SZ]>) {
        let mut tree = MerkleTree::<Sha2_256>::empty(SZ);
        let mut sectors = Vec::with_capacity(num_sects);
        for k in 1..(num_sects + 1) {
            let offset = SZ * (k - 1);
            let sector = [k as u8; SZ];
            sectors.push(sector);
            tree.write(offset, &sector).expect("append sector failed");
        }
        (tree, sectors)
    }

    fn merkle_tree_build_verify_test_case<const SZ: usize>(num_sects: usize) {
        let (tree, sectors) = make_tree_with::<SZ>(num_sects);
        for (k, sector) in sectors.into_iter().enumerate() {
            tree.verify(k * SZ, &sector).expect("verify failed");
        }
    }

    #[test]
    fn merkle_tree_append_verify() {
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(0));
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(1));
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(2));
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(3));
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(4));
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(0) + 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(1) + 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(2) + 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(3) + 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(4) + 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(0) - 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(1) - 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(2) - 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(3) - 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(exp2(4) - 1);
        merkle_tree_build_verify_test_case::<SECTOR_SZ_DEFAULT>(1337);
        merkle_tree_build_verify_test_case::<512>(37);
    }

    #[test]
    fn merkle_tree_data_changed_verify_fails() {
        const SZ: usize = SECTOR_SZ_DEFAULT;
        let mut tree = MerkleTree::<Sha2_256>::empty(SZ);
        let one = [1u8; SZ];
        let mut two = [2u8; SZ];
        let three = [3u8; SZ];
        tree.write(0, &one).expect("append one failed");
        tree.write(SZ, &two).expect("append two failed");
        tree.write(2 * SZ, &three).expect("append three failed");

        two[0] = 7u8;

        tree.verify(0, &one).expect("failed to verify one");
        tree.verify(SZ, &two)
            .expect_err("verify two was expected to fail");
        tree.verify(2 * SZ, &three).expect("failed to verify three");
    }

    #[test]
    fn merkle_tree_root_not_verified_verify_fails() {
        const SZ: usize = SECTOR_SZ_DEFAULT;
        let mut tree = MerkleTree::<Sha2_256>::empty(SZ);
        let one = [1u8; SZ];
        let two = [2u8; SZ];
        let three = [3u8; SZ];
        tree.write(0, &one).expect("append one failed");
        tree.write(SZ, &two).expect("append two failed");
        tree.write(2 * SZ, &three).expect("append three failed");
        let vec = to_vec(&tree).expect("to_vec failed");
        let tree: MerkleTree<Sha2_256> = from_vec(&vec).expect("from_vec failed");

        tree.verify(SZ, &two)
            .expect_err("verify succeeded, though it should have failed");
    }

    fn merkle_stream_sequential_test_case(sect_sz: usize, sect_count: usize) {
        let mut stream = MerkleStream::new(MerkleTree::<Sha2_256>::empty(sect_sz))
            .try_compose(Cursor::new(vec![0u8; sect_count * sect_sz]))
            .expect("compose failed");
        for k in 1..(sect_count + 1) {
            let sector = vec![k as u8; sect_sz];
            stream.write(&sector).expect("write failed");
        }
        stream.seek(SeekFrom::Start(0)).expect("seek failed");
        for k in 1..(sect_count + 1) {
            let expected = vec![k as u8; sect_sz];
            let mut actual = vec![0u8; sect_sz];
            stream.read(&mut actual).expect("read failed");
            assert_eq!(expected, actual);
        }
    }

    #[test]
    fn merkle_stream_sequential() {
        merkle_stream_sequential_test_case(SECTOR_SZ_DEFAULT, 20);
        merkle_stream_sequential_test_case(SECTOR_SZ_DEFAULT, 200);
        merkle_stream_sequential_test_case(SECTOR_SZ_DEFAULT, 800);
        merkle_stream_sequential_test_case(512, 25);
        merkle_stream_sequential_test_case(8192, 20);
    }

    fn merkle_stream_random_test_case(rando: Randomizer, sect_sz: usize, sect_ct: usize) {
        let mut stream = MerkleStream::new(MerkleTree::<Sha2_256>::empty(sect_sz))
            .try_compose(Cursor::new(vec![0u8; sect_sz * sect_ct]))
            .expect("compose failed");
        let indices: Vec<usize> = rando.take(sect_ct).map(|e| e % sect_ct).collect();
        for index in indices.iter().map(|e| *e) {
            let offset = sect_sz * index;
            stream
                .seek(SeekFrom::Start(offset as u64))
                .expect("seek to write failed");
            let sector = vec![index as u8; sect_sz];
            stream.write(&sector).expect("write failed");
        }
        for index in indices.iter().map(|e| *e) {
            let offset = sect_sz * index;
            stream
                .seek(SeekFrom::Start(offset as u64))
                .expect("seek to read failed");
            let expected = vec![index as u8; sect_sz];
            let mut actual = vec![0u8; sect_sz];
            stream.read(&mut actual).expect("read failed");
            assert_eq!(expected, actual);
        }
    }

    #[test]
    fn merkle_stream_random() {
        const SEED: [u8; Randomizer::HASH.len()] = [3u8; Randomizer::HASH.len()];
        merkle_stream_random_test_case(Randomizer::new(SEED), SECTOR_SZ_DEFAULT, 2);
        merkle_stream_random_test_case(Randomizer::new(SEED), SECTOR_SZ_DEFAULT, 4);
        merkle_stream_random_test_case(Randomizer::new(SEED), SECTOR_SZ_DEFAULT, 8);
        merkle_stream_random_test_case(Randomizer::new(SEED), SECTOR_SZ_DEFAULT, 20);
        merkle_stream_random_test_case(Randomizer::new(SEED), SECTOR_SZ_DEFAULT, 200);
        merkle_stream_random_test_case(Randomizer::new(SEED), SECTOR_SZ_DEFAULT, 800);
        merkle_stream_random_test_case(Randomizer::new(SEED), 8192, 63);
    }

    #[test]
    fn compose_merkle_and_secret_streams() {
        const SECT_SZ: usize = 4096;
        const SECT_CT: usize = 16;
        let memory = Cursor::new([0u8; SECT_SZ * SECT_CT]);
        let merkle = MerkleStream::new(MerkleTree::<Sha2_256>::empty(SECT_SZ))
            .try_compose(memory)
            .expect("compose for merkle failed");
        let key = SymKey::generate(SymKeyKind::Aes256Cbc).expect("key generation failed");
        let mut secret = SecretStream::new(key)
            .try_compose(merkle)
            .expect("compose for secret failed");
        let secret_sect_sz = secret.sector_sz();
        write_fill(&mut secret, secret_sect_sz, SECT_CT);
        secret.seek(SeekFrom::Start(0)).expect("seek failed");
        read_check(&mut secret, secret_sect_sz, SECT_CT);
    }

    #[test]
    fn compress_then_encrypt() {
        use brotli::{CompressorWriter, Decompressor};
        const SECT_SZ: usize = 4096;
        const SECT_CT: usize = 16;
        let params = BrotliParams::new(SECT_SZ, 8, 20);
        let key = SymKey::generate(SymKeyKind::Aes256Cbc).expect("key generation failed");
        let mut inner = SectoredBuf::new()
            .try_compose(
                SecretStream::new(key)
                    .try_compose(SectoredCursor::new([0u8; SECT_SZ * SECT_CT], SECT_SZ))
                    .expect("compose for inner failed"),
            )
            .expect("compose with sectored buffer failed");
        {
            let write: CompressorWriter<_> = params
                .clone()
                .try_compose(&mut inner)
                .expect("compose for write failed");
            write_fill(write, SECT_SZ, SECT_CT);
        }
        inner.seek(SeekFrom::Start(0)).expect("seek failed");
        {
            let read: Decompressor<_> = params
                .try_compose(&mut inner)
                .expect("compose for read failed");
            read_check(read, SECT_SZ, SECT_CT);
        }
    }

    /// Tests that validate the dependencies of this module.
    mod dependency_tests {
        use super::*;
        use openssl::{
            ec::{EcGroup, EcKey},
            nid::Nid,
        };

        /// This test validates that data encrypted with AES 256 CBC can later be decrypted.
        #[test]
        fn aes_256_cbc_roundtrip() {
            use super::*;
            let expected = b"We attack at the crack of noon!";
            let cipher = Cipher::aes_256_cbc();

            let ciphertext = openssl_encrypt(cipher, &KEY, Some(&IV), expected).unwrap();
            let actual = openssl_decrypt(cipher, &KEY, Some(&IV), ciphertext.as_slice()).unwrap();

            assert_eq!(expected, actual.as_slice());
        }

        /// Tests that the keys for the SECP256K1 curve are the expected sizes.
        #[test]
        fn secp256k1_key_lengths() {
            let group = EcGroup::from_curve_name(Nid::SECP256K1).unwrap();
            let key = EcKey::generate(&group).unwrap();
            let public = key.public_key_to_der().unwrap();
            let private = key.private_key_to_der().unwrap();
            let public_len = public.len();
            let private_len = private.len();
            assert_eq!(88, public_len);
            assert_eq!(118, private_len);
        }

        #[test]
        fn ed25519_key_lengths() {
            let key = PKey::generate_x25519().unwrap();
            let public = key.public_key_to_der().unwrap();
            let private = key.private_key_to_der().unwrap();
            let public_len = public.len();
            let private_len = private.len();
            assert_eq!(44, public_len);
            assert_eq!(48, private_len);
        }
    }
}
